"""
mock_topics = {
    'sample_topic':{
        1: [],
        2: [],
        3: [],
        4: []
    }
}

offset_store = {
    'sample_topic_1': {
        'first_offset': 0,
        'next_offset': 0,
    }
}
"""
from confluent_kafka import KafkaException
from .message import Message
from copy import deepcopy

mock_topics: dict[str, dict[int, list[Message]]] = {}

offset_store: dict[str, dict[str, int]] = {}


class KafkaStore:
    """
    In memory kafka store
    """
    FIRST_OFFSET = 'first_offset'
    NEXT_OFFSET = 'next_offset'

    def __init__(self, clean: bool = False):
        if clean:
            mock_topics.clear()
            offset_store.clear()

    @staticmethod
    def is_topic_exist(topic: str) -> bool:
        return topic in mock_topics.keys()

    @classmethod
    def is_partition_exist_on_topic(cls, topic: str, partition_num: int) -> bool:
        if not cls.is_topic_exist(topic=topic):
            raise KafkaException('Topic Does not exist')

        return mock_topics[topic].get(partition_num, None) is not None

    @staticmethod
    def get_number_of_partition(topic: str) -> int:
        return len(mock_topics[topic].keys())

    @staticmethod
    def create_topic(topic: str):
        if mock_topics.get(topic, None) is not None:
            raise KafkaException(f'{topic} exist is fake kafka')

        mock_topics[topic] = {}

    def create_partition(self, topic: str, partitions: int):
        if not self.is_topic_exist(topic=topic):
            self.create_topic(topic=topic)

        len_of_current_partition = len(mock_topics[topic].keys())
        if partitions >= len_of_current_partition:
            for i in range(len_of_current_partition, partitions):
                mock_topics[topic][i] = []
                offset_store[self.get_offset_store_key(topic, i)] = {
                    self.FIRST_OFFSET: 0,
                    self.NEXT_OFFSET: 0
                }

        else:
            raise KafkaException('can not decrease partition of topic')

    def remove_topic(self, topic: str):
        if not self.is_topic_exist(topic=topic):
            return

        mock_topics.pop(topic)

        offset_store_keys = deepcopy(list(offset_store.keys()))
        for offset_key in offset_store_keys:
            if topic in offset_key:
                offset_store.pop(offset_key)

    def set_first_offset(self, topic: str, partition: int, value: int):
        offset_store_key = self.get_offset_store_key(topic=topic, partition=partition)
        first_offset = self.get_partition_first_offset(topic=topic, partition=partition)
        next_offset = self.get_partition_next_offset(topic=topic, partition=partition)

        if first_offset < value <= next_offset:
            offset_store[offset_store_key][self.FIRST_OFFSET] = value

    def _add_next_offset(self, topic: str, partition: int):
        offset_store_key = self.get_offset_store_key(topic=topic, partition=partition)
        offset_store[offset_store_key][self.NEXT_OFFSET] += 1

    def get_offset_store_key(self, topic: str, partition: int):
        return f'{topic}*{partition}'

    def produce(self, message: Message, topic: str, partition: int):
        if not self.is_topic_exist(topic=topic):
            raise KafkaException(f'can not produce on {topic}, Topic Does Not Exist')

        if mock_topics[topic].get(partition, None) is None:
            raise KafkaException(f'can not produce on partition {partition} of {topic}, partition does not exist')

        if partition is None:
            raise KafkaException('you must assigne partition when you want to produce message')

        # add message to topic
        mock_topics[topic][partition].append(message)

        self._add_next_offset(topic=topic, partition=partition)

    def get_message(self, topic: str, partition: int, offset: int) -> Message:
        return self.get_messages_in_partition(topic=topic, partition=partition)[offset]

    def get_partition_first_offset(self, topic: str, partition: int) -> int:
        offset_store_key = self.get_offset_store_key(topic=topic, partition=partition)
        return offset_store[offset_store_key][self.FIRST_OFFSET]

    def get_partition_next_offset(self, topic: str, partition: int) -> int:
        offset_store_key = self.get_offset_store_key(topic=topic, partition=partition)
        return offset_store[offset_store_key][self.NEXT_OFFSET]

    @staticmethod
    def topic_list() -> list[str]:
        return list(mock_topics.keys())

    @staticmethod
    def partition_list(topic: str) -> list[int]:
        return list(mock_topics[topic].keys())

    @staticmethod
    def get_messages_in_partition(topic: str, partition: int) -> list[Message]:
        return mock_topics[topic][partition]

    def number_of_message_in_topic(self, topic: str) -> int:
        count_of_messages = 0
        for partition in self.partition_list(topic=topic):
            count_of_messages += len(self.get_messages_in_partition(topic=topic, partition=partition))

        return count_of_messages

    def clear_topic_messages(self, topic: str):
        for partition in self.partition_list(topic=topic):
            self.clear_partition_messages(topic=topic, partition=partition)

    @staticmethod
    def clear_partition_messages(topic: str, partition: int):
        mock_topics[topic][partition] = []

    def reset_offset(self, topic: str, strategy: str = 'latest'):
        for partition in self.partition_list(topic=topic):
            key = self.get_offset_store_key(topic, partition)

            if strategy == 'latest':
                offset_store[key][self.FIRST_OFFSET] = offset_store[key][self.NEXT_OFFSET]

            elif strategy == 'earliest':
                offset_store[key][self.FIRST_OFFSET] = 0

    @staticmethod
    def fresh():
        mock_topics.clear()
        offset_store.clear()



from functools import wraps
from mockafka import FakeProducer


def produce(**decorator_args):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Extract parameters from the decorator_args
            topic = decorator_args.get('topic', None)
            value = decorator_args.get('value', None)
            key = decorator_args.get('key', None)
            partition = decorator_args.get('partition', None)

            fake_producer = FakeProducer()

            fake_producer.produce(topic=topic, partition=partition, value=value, key=key)
            result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator




from typing import Optional

from confluent_kafka import KafkaError


class Message:
    def __init__(self, *args, **kwargs):
        self._headers: Optional[dict] = kwargs.get('headers', None)
        self._key: Optional[str] = kwargs.get('key', None)
        self._value: Optional[str] = kwargs.get('value', None)
        self._topic: Optional[str] = kwargs.get('topic', None)
        self._offset: Optional[int] = kwargs.get('offset', None)
        self._error: Optional[KafkaError] = kwargs.get('error', None)
        self._latency: Optional[float] = kwargs.get('latency', None)
        self._leader_epoch: Optional[int] = kwargs.get('leader_epoch', None)
        self._partition: Optional[int] = kwargs.get('partition', None)
        self._timestamp: int = kwargs.get('timestamp', None)

    def offset(self, *args, **kwargs):
        return self._offset

    def latency(self, *args, **kwargs):
        return self._latency

    def leader_epoch(self, *args, **kwargs):
        return self._leader_epoch

    def headers(self, *args, **kwargs):
        return self._headers

    def key(self, *args, **kwargs):
        return self._key

    def value(self, *args, **kwargs):
        return self._value

    def timestamp(self, *args, **kwargs):
        return self._timestamp

    def topic(self, *args, **kwargs):
        return self._topic

    def value(self, payload):
        return self._value

    def error(self):
        return self._error

    def set_headers(self, *args, **kwargs):  # real signature unknown
        pass

    def set_key(self, *args, **kwargs):  # real signature unknown
        pass

    def set_value(self, *args, **kwargs):  # real signature unknown
        pass



class BrokerMetadata(object):
    """
    Provides information about a Kafka broker.

    This class is typically not user instantiated.
    """

    def __init__(self):
        self.id = 1
        """Broker id"""
        self.host = 'fakebroker'
        """Broker hostname"""
        self.port = 9091
        """Broker port"""

    def __repr__(self):
        return "BrokerMetadata({}, {}:{})".format(self.id, self.host, self.port)

    def __str__(self):
        return "{}:{}/{}".format(self.host, self.port, self.id)



from mockafka.cluster_metadata import ClusterMetadata
from mockafka.kafka_store import KafkaStore
from mockafka.message import Message

__all__ = ["FakeProducer"]


class FakeProducer(object):
    def __init__(self, config: dict = None):
        self.kafka = KafkaStore()

    def produce(self, topic, value=None, *args, **kwargs):
        # create a message and call produce kafka
        message = Message(value=value, topic=topic, *args, **kwargs)
        self.kafka.produce(message=message, topic=topic, partition=kwargs['partition'])

    def list_topics(self, topic=None, *args, **kwargs):
        return ClusterMetadata(topic)

    def abort_transaction(self, timeout=None):
        # This method Does not support in mockafka
        pass

    def begin_transaction(self):
        # This method Does not support in mockafka
        pass

    def commit_transaction(self, timeout=None):
        # This method Does not support in mockafka
        pass

    def flush(self, timeout=None):
        # This method Does not support in mockafka
        return 0

    def init_transactions(self, timeout=None):
        # This method Does not support in mockafka
        pass

    def poll(self, timeout=None):
        # This method Does not support in mockafka
        return 0

    def purge(self, in_queue=True, *args, **kwargs):
        # This method Does not support in mockafka
        pass

    def send_offsets_to_transaction(self, positions, group_metadata,
                                    timeout=None):
        # This method Does not support in mockafka
        pass



from functools import wraps
from mockafka import FakeConsumer


def consume(topics: list[str], auto_commit: bool = True):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Extract parameters from the decorator_args

            fake_consumer = FakeConsumer()
            fake_consumer.subscribe(topics=topics)

            # not complete yet
            # while True:
            #     message = fake_consumer.poll()

            result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator




from functools import wraps
from mockafka import FakeProducer


def bulk_produce(list_of_messages: list[dict[str, str]]):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            fake_producer = FakeProducer()
            # Extract parameters from the decorator_args
            for item in list_of_messages:
                value = item.get('value', None)
                key = item.get('key', None)
                topic = item.get('topic', None)
                partition = item.get('partition', None)
                timestamp = item.get('timestamp', None)
                headers = item.get('headers', None)

                fake_producer.produce(topic=topic, partition=partition, value=value, key=key,
                                      timestamp=timestamp, headers=headers)

            result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator



from time import sleep

from mockafka.cluster_metadata import ClusterMetadata
from mockafka.message import Message
from mockafka.kafka_store import KafkaStore, KafkaException

__all__ = ["FakeConsumer"]


class FakeConsumer(object):
    def __init__(self, *args, **kwargs):
        self.kafka = KafkaStore()
        self.consumer_store = {}
        self.subscribed_topic: list = []

    def consume(self, num_messages=1, *args, **kwargs):
        return self.poll()

    def close(self, *args, **kwargs):
        self.consumer_store = {}
        self.subscribed_topic = []

    def commit(self, message: Message = None, *args, **kwargs):
        if message:
            # we can keep (topic, partition, offset) in a message when we produce
            # and commit it by changing offset of topic i did not implement it yet
            pass

        else:
            for item in self.consumer_store:
                topic, partition = item.split('*')
                if self.kafka.get_partition_first_offset(topic, partition) <= self.consumer_store[item]:
                    self.kafka.set_first_offset(topic=topic, partition=partition, value=self.consumer_store[item])

            self.consumer_store = {}

    def list_topics(self, topic=None, *args, **kwargs):
        return ClusterMetadata(topic=topic)

    def poll(self, timeout=None):
        if timeout:
            sleep(timeout)

        for topic in self.subscribed_topic:
            for partition in self.kafka.partition_list(topic=topic):
                first_offset = self.kafka.get_partition_first_offset(topic=topic, partition=partition)
                next_offset = self.kafka.get_partition_next_offset(topic=topic, partition=partition)
                consumer_amount = self.consumer_store.get(self._get_key(topic, partition))
                if first_offset == next_offset:
                    continue

                if consumer_amount == next_offset:
                    continue

                if consumer_amount is not None:
                    self.consumer_store[self._get_key(topic, partition)] += 1
                else:
                    self.consumer_store[self._get_key(topic, partition)] = first_offset + 1

                return self.kafka.get_message(topic=topic, partition=partition, offset=first_offset)

        return None

    def _get_key(self, topic, partition) -> str:
        return f'{topic}*{partition}'

    def subscribe(self, topics, on_assign=None, *args, **kwargs):
        for topic in topics:
            if not self.kafka.is_topic_exist(topic):
                raise KafkaException(f'{topic} Does not exist in kafka.')

            if topic not in self.subscribed_topic:
                self.subscribed_topic.append(topic)

    def unsubscribe(self, *args, **kwargs):
        topics = kwargs['topics']
        for topic in topics:
            if topic in self.subscribed_topic:
                self.subscribed_topic.remove(topic)

    def assign(self, partitions):
        # This method Does not support in mockafka
        pass

    def unassign(self, *args, **kwargs):
        # This method Does not support in mockafka
        pass

    def assignment(self, *args, **kwargs):
        # This method Does not support in mockafka
        return []

    def committed(self, partitions, timeout=None):
        # This method Does not support in mockafka
        return []

    def get_watermark_offsets(self, partition, timeout=None, *args, **kwargs):
        # This method Does not support in mockafka
        return (0, 0)

    def offsets_for_times(self, partitions, timeout=None):
        # This method Does not support in mockafka
        return []

    def pause(self, partitions):
        # This method Does not support in mockafka
        return None

    def position(self, partitions):
        # This method Does not support in mockafka
        return []

    def resume(self, partitions):
        # This method Does not support in mockafka
        return None

    def seek(self, partition):
        # This method Does not support in mockafka
        pass

    def store_offsets(self, message=None, *args, **kwargs):
        # This method Does not support in mockafka
        return None

    def consumer_group_metadata(self):
        # This method Does not support in mockafka
        pass

    def incremental_assign(self, partitions):
        # This method Does not support in mockafka
        pass

    def incremental_unassign(self, partitions):
        # This method Does not support in mockafka
        pass



from confluent_kafka.cimpl import NewTopic, NewPartitions

from mockafka.cluster_metadata import ClusterMetadata
from mockafka.kafka_store import KafkaStore

__all__ = ["FakeAdminClientImpl"]


class FakeAdminClientImpl:
    """
    Kafka Fake Admin Client

    .. py:function:: Admin(**kwargs)

      Create a new AdminClient instance using the provided configuration dict.

    This class should not be used directly, use confluent_kafka.AdminClient
    .
    .. py:function:: len()

      :returns: Number Kafka protocol requests waiting to be delivered to, or returned from, broker.
      :rtype: int
    """

    def __init__(self, clean: bool = True, *args, **kwargs):
        self.kafka = KafkaStore(clean=clean)

    def create_partitions(self, partitions: list[NewPartitions]):
        for partition in partitions:
            self.create_partition(partition)

    def create_partition(self, partition: NewPartitions):
        self.kafka.create_partition(topic=partition.topic, partitions=partition.new_total_count)

    def create_topics(self, topics: list[NewTopic]):
        for topic in topics:
            self.create_topic(topic=topic)

    def create_topic(self, topic: NewTopic):
        self.kafka.create_topic(topic.topic)
        self.create_partitions([NewPartitions(topic.topic, topic.num_partitions)])

    def delete_topics(self, topics, future=None, request_timeout=None,
                      operation_timeout=None):
        for topic in topics:
            self.delete_topic(topic=topic)

    def delete_topic(self, topic: NewTopic):
        self.kafka.remove_topic(topic=topic.topic)

    def describe_acls(self, acl_binding_filter, future,
                      request_timeout=None):
        # This method Does not support in mockafka
        pass

    def describe_configs(self, resources, future, request_timeout=None,
                         broker=None):
        # This method Does not support in mockafka
        pass

    def delete_acls(self, acl_binding_filters, future,
                    request_timeout=None):
        # This method Does not support in mockafka
        pass

    def alter_configs(self, *args, **kwargs):
        # This method Does not support in mockafka
        pass

    def create_acls(self, *args, **kwargs):
        # This method Does not support in mockafka
        pass

    def list_groups(self, group=None, *args,
                    **kwargs):
        # This method Does not support in mockafka
        pass

    def list_topics(self, topic=None, *args,
                    **kwargs):  # real signature unknown; NOTE: unreliably restored from __doc__
        return ClusterMetadata(topic)

    def poll(self, timeout=None):
        # This method Does not support in mockafka
        pass

    def __len__(self, *args, **kwargs):
        pass



from .admin_client import FakeAdminClientImpl
from .producer import FakeProducer
from .conumser import FakeConsumer
from .message import Message
from .decorators import produce, bulk_produce

__all__ = [
    "FakeProducer",
    "FakeConsumer",
    "FakeAdminClientImpl",
    "Message",
    "produce",
    "bulk_produce"
]



from mockafka.broker_metadata import BrokerMetadata
from mockafka.kafka_store import KafkaStore
from mockafka.topic_metadata import TopicMetadata


class ClusterMetadata(object):
    """
    Provides information about the Kafka cluster, brokers, and topics.
    Returned by list_topics().

    This class is typically not user instantiated.
    """

    def __init__(self, topic: str = None):
        self.kafka = KafkaStore()
        self.cluster_id = 'eAvlnr_4QISNbc5bIwBRVA'
        self.controller_id = 1
        self.brokers = {1: BrokerMetadata()}
        self.topics = {}
        if topic:
            if self.kafka.is_topic_exist(topic=topic):
                self.topics[topic] = TopicMetadata(topic, self.kafka.partition_list(topic=topic))

        else:
            for topic in self.kafka.topic_list():
                self.topics[topic] = TopicMetadata(topic, self.kafka.partition_list(topic=topic))

        self.orig_broker_id = -1
        self.orig_broker_name = None

    def __repr__(self):
        return "ClusterMetadata({})".format(self.cluster_id)

    def __str__(self):
        return str(self.cluster_id)
